/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2007-2009 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */
/*!
 *  \file   toppers/macro_processor.hpp
 *  \brief  }N̂߂̐錾`
 *
 *  ̃t@CŒ`NX
 *  \code
 *  class macro_processor;
 *  \endcode
 */
#ifndef TOPPERS_MACRO_PROCESSOR_HPP_
#define TOPPERS_MACRO_PROCESSOR_HPP_

#include <string>
#include <vector>
#include <stack>
#include <map>
#include "toppers/workaround.hpp"
#include "toppers/output_file.hpp"
#include "toppers/text.hpp"

namespace toppers
{

  class macro_processor
  {
  public:
    struct context;

    /*!
     *  \struct element macro_processor.hpp "toppers/macro_processor.hpp"
     *  \brief  }NŎgpϐ̒l
     *
     *  mɂ́Aϐ̒l͏ɏtXgƂĈ܂Belement ͏tXg 1 vf\܂B
     */
    struct element
    {
      boost::optional< std::tr1::int64_t > i;
      std::string v;
      std::string s;
    };
    /*!
     *  \brief  }NŎgpϐ\^
     */
    typedef std::vector< element > var_t;

    /*!
     *  \struct func_t macro_processor.hpp "toppers/macro_processor.hpp"
     *  \brief  }Nň֐\^
     *  \note   XPȍ\̂ł߁A݊̂߁AŴƓێĂB
     *  \attention  node ̏Yɗv
     */
    struct func_t
    {
      std::string name;   //!< ֐
      var_t ( * f )( text_line const& line, std::vector< var_t > const&, context* );  //!< e
      void const* node;
    };

    typedef void ( * hook_t )( text_line const&, std::string const&, var_t const&, var_t const&, context* );

    struct context
    {
      std::stack< var_t > stack;
      std::map< std::string, var_t > var_map;
      std::map< std::string, func_t > func_map;
      bool in_function;
      output_file target_file;
      hook_t hook_on_assign;
      text_line line;
      explicit context( hook_t hoa = 0 ) : hook_on_assign( hoa ), in_function( false ) {}
    };

    explicit macro_processor( hook_t hook_on_assign = 0 );
    macro_processor( macro_processor const& other );
    explicit macro_processor( text const& in, hook_t hook_on_assign = 0 );
    ~macro_processor();
    macro_processor& operator=( macro_processor const& other );
    void swap( macro_processor& other );
    void evaluate( text const& in );
    void set_var( std::string const& name, var_t const& value );
    void set_var( std::string const& name, long name2, var_t const& value );
    var_t const& get_var( std::string const& name ) const;
    var_t const& get_var( std::string const& name, long name2 ) const;
    void add_builtin_function( func_t const& f );

    static void remove_comment( text const& in, text& out );
    static void expand_include( text const& in, text& out );
    static void preprocess( text const& in, text& out );
    static std::tr1::int64_t to_integer( var_t const& var, context const* p_ctx );
    static std::string to_string( var_t const& var, context const* p_ctx );
    static bool check_arity( text_line const& line, std::size_t arity, std::size_t valid, char const* function_name );
    static var_t call_user_function( text_line const& line, std::vector< var_t > const& arg_list, context* p_ctx );
  private:
    context* p_ctx_;
    static func_t const builtin_function_table[];
  };

  inline bool operator == ( macro_processor::element const& lhs, macro_processor::element const& rhs )
  {
    return lhs.i == rhs.i;
  }

  inline bool operator < ( macro_processor::element const& lhs, macro_processor::element const& rhs )
  {
    return lhs.i < rhs.i;
  }

  struct element_eq
  {
    explicit element_eq( std::tr1::int64_t value ) : value_( value ) {}
    bool operator()( toppers::macro_processor::element const& e ) const
    {
      return e.i && ( e.i.get() == value_ );
    }
    std::tr1::int64_t value_;
  };

}

#endif  // ! TOPPERS_MACRO_PROCESSOR_HPP_
